#!/usr/bin/env python
"""Default bootstrap script.

Usage:
  sudo bootstrap.py [filepattern ...] [--jail-extra-files filepattern ...]
      [--jail-dir path] [--min-uid uid] [--min-gid gid] [--script script]

  For example:
    $> sudo bootstrap.py ~/code/mywebapp
    $> sudo bootstrap.py ~/code/mywebapp --jail-extra-files /usr/bin
       --jail-dir /jails/jail --min-uid 61015 --min-gid 61015 --script script.sh

  Receives a set of filepatterns that are the set of python files that make
  the webapp. This can be listed explicitly, or a directory can be specified.
  (Defaults to the current directory if none is specified.)

  All the files that are part of the webapp will be copied to the jail, in
  addition to these you can specify extra files that should be copied to the
  jail with the --jail-extra-files flag. (sys.path are already copied in.)

  The jail-dir is the location where to put the jail. (Defaults to /jail)

  The min-uid and min-gid respectively specify the starting points of uids and
  gids that are not used by the user so that detected services can run with
  independent and unused uids and gids. (Both default to 1000)

  The script is a file that will be executed after files are moved to the
  jail but before service processes are launched so the user can perform
  additional actions in the right environment, such as creating database files.
"""

import argparse
import lib


def main():
  parser = argparse.ArgumentParser(
      description=(
          'Run using sudo. Bootstraps the environment: starts all pspy '
          'services'))
  parser.add_argument(
      'app_files', nargs='*', default=lib.DEFAULT_CONFIG.app_files, help=(
          'set of filepatterns that make the webapp. These can be listed '
          'explicitly or a directory can be specified. Defaults to the cwd'))
  parser.add_argument(
      '--jail-extra-files', nargs='+',
      default=lib.DEFAULT_CONFIG.jail_extra_files, help=(
          'set of extra filepatterns that should be copied to the jail (the '
          'app files and sys.path are always copied). Defaults to no extra '
          'files'))
  parser.add_argument('--jail-dir', default=lib.DEFAULT_CONFIG.jail_dir, help=(
      'the location where to put the jail. Defaults to /jail'))
  parser.add_argument('--min-uid', default=lib.DEFAULT_CONFIG.min_uid, help=(
      'the starting point of uids that are not used by the user. Defaults to '
      '1000'))
  parser.add_argument('--min-gid', default=lib.DEFAULT_CONFIG.min_gid, help=(
      'the starting point of gids that are not used by the user. Defaults to '
      '1000'))
  parser.add_argument('--script', default='', help=(
      'a shell script to run after the application has been jailed'))

  args = parser.parse_args()
  lib.bootstrap(config=lib.Config(
      args.jail_dir, args.app_files, args.jail_extra_files, args.min_uid, 
      args.min_gid, args.script))


if __name__ == '__main__':
  main()
