import collections
import errno
import os
import parser
import shutil
import subprocess
import sys


def parse_app(fps):
  """Parses the app, returns the pspy services and related file permissions."""
  py_files = []
  for fp in fps:
    if os.path.isdir(fp):
      for dirpath, _, filenames in os.walk(fp):
        py_files.extend([
            os.path.join(dirpath, f) for f in filenames if f.endswith('.py')])
    elif fp.endswith('.py'):
      py_files.append(fp)
  ret = {}
  # We aggregate the results from each file together into ret.
  for py_file in py_files:
    service_prov = parser.parse_file(py_file)
    for prov in service_prov:
      if prov in ret:
        ret[prov].extend(service_prov[prov])
      else:
        ret[prov] = service_prov[prov]
  return ret


Config = collections.namedtuple(
    'Config', [
        'jail_dir', 'app_files', 'jail_extra_files', 'min_uid', 'min_gid', 'script'])
DEFAULT_CONFIG = Config('/jail', [os.getcwd()], [], 1000, 1000, '')


def bootstrap(config=DEFAULT_CONFIG, debug_writer=sys.stdout.write):
  """Bootstraps the pspy environment.
 
  1. creates the jail directory;
  2. copies the app_files and jail_extra_files to the jail;
  3. parses the app files looking for pspy services;
  4. starts the service providers that provide the services detected.

  Along the way it will print useful status information to debug_writer.
  """
  # Some helpers for outputting info.
  dw = lambda status: lambda msg: debug_writer('[%s] %s\n' % (status, msg))
  dw_fail = dw('FAILURE')
  dw_success = dw('SUCCESS')
  dw_info = dw('INFO')

  if os.geteuid() != 0:
    dw_fail('You need to execute bootstrap with root privileges')
    return

  def safe_copy(files):
    """Copies the given files to the jail."""
    files.sort()
    for f in files:
      jail_dst = os.path.join(
          config.jail_dir, f if not os.path.isabs(f) else f[1:])
      try:
        shutil.copytree(f, jail_dst)
      except OSError as e:
        if e.errno == errno.ENOTDIR:
          dir_path = os.path.split(jail_dst)[0]
          if not os.path.exists(dir_path):
            os.makedirs(os.path.split(jail_dst)[0])
          shutil.copy(f, jail_dst)
        elif e.errno == errno.EEXIST or e.errno == errno.ENOENT:
          continue
        else:
          dw_fail('Unable to copy file %s' % f)
          raise
      except shutil.Error as e:
        continue
      dw_info('...copied %s to %s' % (f, jail_dst))

  dw_info('Creating jail directory at %s' % config.jail_dir)
  try:
    if os.path.exists(config.jail_dir):
      dw_info('Removing existing dir at directory %s' % config.jail_dir)
      shutil.rmtree(config.jail_dir)
    os.mkdir(config.jail_dir)
  except:
    dw_fail('Unable to create jail directory at %s' % config.jail_dir)
    raise
  dw_success('Created jail directory at %s' % config.jail_dir)
  safe_copy(['/System/Library/Frameworks/SystemConfiguration.framework/'])
  dw_info('Copying app files to jail')
  safe_copy(config.app_files)
  dw_success('Copied app to jail')
  dw_info('Copying extra files to jail')
  safe_copy(sys.path)
  safe_copy(config.jail_extra_files)
  dw_success('Copied extra files to jail')

  if config.script != '':
    if not os.path.exists(config.script):
      dw_failure('Script %s not found' % config.script)
      return
    dw_info('Executing script %s' % config.script)
    subprocess.Popen(
        config.script, stdin=sys.stdin, stdout=sys.stdout,
        shell=True).wait()
    dw_info('Done executing script %s' % config.script)

  dw_info('Parsing app code to detect pspy services')
  service_providers = parse_app(config.app_files)
  dw_info('Done detecting services')
  uid = int(config.min_uid)
  next_gid = int(config.min_gid)
  gids_dict = {}
  for service_provider in service_providers:
    dw_info('Preparing to launch service provider %s' % service_provider)
    dw_info('Creating socket for service provider %s' % service_provider)
    suffix = service_provider
    if '.' in service_provider:
      suffix = service_provider.rsplit('.', 1)[1]
    socket_path = os.path.join(config.jail_dir, 'sock-%s' % suffix)
    os.mkdir(socket_path)  # 0777
    services, group_name, permissions, add_group_names = service_providers[
        service_provider]
    next_gid, gids_dict, gid, groups = get_gids(next_gid, gids_dict, group_name, add_group_names)
    os.chown(socket_path, uid, gid)
    dw_success('Created socket dir %s' % socket_path)
    dw_info('Setting file permissions for this service provider')
    for p in permissions:
      path = os.path.join(config.jail_dir, p if not os.path.isabs(p) else p[1:])
      recur_perms(path, uid, gid, permissions[p])
    dw_info('Launching service provider %s' % service_provider)
    dw_info(
        'Service provider %s provides services: %s' % (
            service_provider, services))
    subprocess.Popen(
        ('pspy-rpc-server --service-provider %s --services %s '
         '--socket %s --jail-dir %s --uid %s --gid %s '
         '--additional-groups %s') % (
            suffix, ' '.join(x for x in services),
            socket_path[len(config.jail_dir):], config.jail_dir, uid, gid,
            ' '.join(str(x) for x in groups)),
        shell=True)
    dw_success('Launched service provider %s' % service_provider)
    uid += 1


def get_gids(next_gid, gids_dict, group_name, additional_group_names):
  """Determine the gids for the given names based on id availability."""
  if group_name in gids_dict:
    gid = gids_dict[group_name]
  else:
    gid = next_gid
    next_gid += 1
    if group_name != '':
      gids_dict[group_name] = gid
  groups = []
  for name in additional_group_names:
    if name in gids_dict:
      groups.append(gids_dict[name])
    else:
      groups.append(next_gid)
      gids_dict[name] = next_gid
      next_gid += 1
  return (next_gid, gids_dict, gid, groups)


def recur_perms(path, uid, gid, perm):
  """Recursively set ownership and permissions on 'path'."""
  os.chown(path, uid, gid)
  os.chmod(path, perm)
  if os.path.isdir(path):
    for item in os.listdir(path):
      itempath = os.path.join(path, item)
      recur_perms(itempath, uid, gid, perm)
